﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace SunBigUInteger.SunBigUInt
{
    public abstract class SunBigUIntAbstract<TClass> where TClass : SunBigUIntAbstract<TClass>
    {
        public SunBigUIntAbstract() {  }

        public abstract TClass FromLong(ulong number);
        public abstract TClass FromString(string number);
        public abstract TClass FromHexString(string number);

        /* Addition */
        public static TClass operator +(SunBigUIntAbstract<TClass> c1, SunBigUIntAbstract<TClass> c2)
        {
            return c1.Add(c2);
        }
        public abstract TClass Add(SunBigUIntAbstract<TClass> c1);

        /* Subtraction */
        public static TClass operator -(SunBigUIntAbstract<TClass> c1, SunBigUIntAbstract<TClass> c2)
        {
            return c1.Sub(c2);
        }
        public abstract TClass Sub(SunBigUIntAbstract<TClass> c1);

        /* Multiplication */
        public static TClass operator *(SunBigUIntAbstract<TClass> c1, SunBigUIntAbstract<TClass> c2)
        {
            return c1.Mul(c2);
        }
        public abstract TClass Mul(SunBigUIntAbstract<TClass> c1);


        /* Division */
        public static TClass operator /(SunBigUIntAbstract<TClass> c1, SunBigUIntAbstract<TClass> c2)
        {
            return c1.Div(c2);
        }
        public abstract TClass Div(SunBigUIntAbstract<TClass> c1);

        public abstract void DivRem(SunBigUIntAbstract<TClass> a, SunBigUIntAbstract<TClass> b, out SunBigUIntAbstract<TClass> q, out SunBigUIntAbstract<TClass> r);
        public abstract TClass DivRem(SunBigUIntAbstract<TClass> a, SunBigUIntAbstract<TClass> b, out SunBigUIntAbstract<TClass> r);
        /* Modulus */
        public abstract TClass Mod(SunBigUIntAbstract<TClass> c1);

        /* Pow / ModPow */
        public abstract TClass Pow(SunBigUIntAbstract<TClass> a, SunBigUIntAbstract<TClass> b);
        public abstract TClass Pow(SunBigUIntAbstract<TClass> a, int b);
        public abstract TClass ModPow(SunBigUIntAbstract<TClass> value, SunBigUIntAbstract<TClass> exponent, SunBigUIntAbstract<TClass> modulus);
        public abstract TClass ModPow(int value, int exponent, int modulus);

        /* Increment / Decrement */
        public static TClass operator ++(SunBigUIntAbstract<TClass> a)
        {
            return a.Inc(a);
        }
        public abstract TClass Inc(SunBigUIntAbstract<TClass> a);

        public static TClass operator --(SunBigUIntAbstract<TClass> a)
        {
            return a.Dec(a);
        }
        public abstract TClass Dec(SunBigUIntAbstract<TClass> a);

        /* Relational operators */
        public static bool operator >(SunBigUIntAbstract<TClass> a, SunBigUIntAbstract<TClass> b)
        {
            return a.Greater(b);
        }
        public abstract bool Greater(SunBigUIntAbstract<TClass> a);

        public static bool operator >=(SunBigUIntAbstract<TClass> a, SunBigUIntAbstract<TClass> b)
        {
            return a.GreaterEqual(b);
        }
        public abstract bool GreaterEqual(SunBigUIntAbstract<TClass> a);

        public static bool operator <(SunBigUIntAbstract<TClass> a, SunBigUIntAbstract<TClass> b)
        {
            return a.Less(b);
        }
        public abstract bool Less(SunBigUIntAbstract<TClass> a);

        public static bool operator <=(SunBigUIntAbstract<TClass> a, SunBigUIntAbstract<TClass> b)
        {
            return a.LessEqual(b);
        }
        public abstract bool LessEqual(SunBigUIntAbstract<TClass> a);

        public static bool operator ==(SunBigUIntAbstract<TClass> a, SunBigUIntAbstract<TClass> b)
        {
            return a.Equal(b);
        }
        public abstract bool Equal(SunBigUIntAbstract<TClass> a);

        public static bool operator !=(SunBigUIntAbstract<TClass> a, SunBigUIntAbstract<TClass> b)
        {
            return a.Unequal(b);
        }
        public abstract bool Unequal(SunBigUIntAbstract<TClass> a);


        /* Even / Odd */
        public abstract bool IsEven();
        public abstract bool IsOdd();

        public override bool Equals(object obj)
        {
            if (obj.GetType() != this.GetType())
            {
                return false;
            }

            return this.Equals(obj as TClass);
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }


        //{
        //    if (obj.GetType() != this.GetType())
        //    {
        //        return false;
        //    }
        //
        //    return this.Equals(obj as SunBigUIntAbstract<TClass>);
        //}
        //
        //public bool Equals(SunBigUInt obj)
        //{
        //    if (obj is null)
        //    {
        //        return false;
        //    }
        //
        //    if (Object.ReferenceEquals(this, obj))
        //    {
        //        return true;
        //    }
        //
        //    return this == obj;
        //}
        //
        //public override int GetHashCode()
        //{
        //    return base.GetHashCode();
        //}
    }
}
