﻿using SunBigUInteger.SunBigUInt;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using System.Text;
using System.Threading.Tasks;

namespace SunBigUInteger.SunBigUIntTests
{
    class SunBigUInt10SpecificTests
    {
        private readonly ITestHandler testHandler;
        public int NumOfFailedTestcases { private set; get; }

        public SunBigUInt10SpecificTests(ITestHandler ts)
        {
            this.testHandler = ts;
            NumOfFailedTestcases = 0;
        }

        public void Run()
        {
            RunMultplicationSimpleTestCases();
            RunLeftShiftByBaseTestCases();
            RunRightShiftByBaseTestCases();
            RunLeftShiftByBinaryTestCases();
            RunRightShiftByBinaryTestCases();
            RunPowSimpleTestCases();
        }

        public void RunMultplicationSimpleTestCases()
        {
            testHandler.TestGroupStart("Multipication simple (*)");

            AssertEqual(new BigInteger(0) * new BigInteger(0), SunBigUInt10.SFromLong(0).MulSimple(SunBigUInt10.SFromLong(0)));
            AssertEqual(new BigInteger(0) * new BigInteger(19), SunBigUInt10.SFromLong(0).MulSimple(SunBigUInt10.SFromLong(19)));
            AssertEqual(new BigInteger(5643) * new BigInteger(0), SunBigUInt10.SFromLong(5643).MulSimple(SunBigUInt10.SFromLong(0)));
            AssertEqual(new BigInteger(1) * new BigInteger(378), SunBigUInt10.SFromLong(1).MulSimple(SunBigUInt10.SFromLong(378)));
            AssertEqual(new BigInteger(235409990009) * new BigInteger(1), SunBigUInt10.SFromLong(235409990009).MulSimple(SunBigUInt10.SFromLong(1)));
            AssertEqual(new BigInteger(109) * new BigInteger(45), SunBigUInt10.SFromLong(109).MulSimple(SunBigUInt10.SFromLong(45)));
            AssertEqual(new BigInteger(45) * new BigInteger(109), SunBigUInt10.SFromLong(45).MulSimple(SunBigUInt10.SFromLong(109)));
            AssertEqual(new BigInteger(99108) * new BigInteger(3448709199), SunBigUInt10.SFromLong(99108).MulSimple(SunBigUInt10.SFromLong(3448709199)));
            AssertEqual(new BigInteger(9) * new BigInteger(134), SunBigUInt10.SFromLong(9).MulSimple(SunBigUInt10.SFromLong(134)));
            AssertEqual(new BigInteger(134) * new BigInteger(9), SunBigUInt10.SFromLong(134).MulSimple(SunBigUInt10.SFromLong(9)));
            AssertEqual(new BigInteger(99) * new BigInteger(8765), SunBigUInt10.SFromLong(99).MulSimple(SunBigUInt10.SFromLong(8765)));
            AssertEqual(new BigInteger(8765) * new BigInteger(99), SunBigUInt10.SFromLong(8765).MulSimple(SunBigUInt10.SFromLong(99)));

            testHandler.TestGroupEnd();
        }

        public void RunLeftShiftByBaseTestCases()
        {
            testHandler.TestGroupStart("Left shift (by base)");

            AssertEqual(123, SunBigUInt10.LeftShiftByBase(123, 0));
            AssertEqual(0, SunBigUInt10.LeftShiftByBase(0, 3));
            AssertEqual(1230, SunBigUInt10.LeftShiftByBase(123, 1));
            AssertEqual(12300000000, SunBigUInt10.LeftShiftByBase(123, 8));
            AssertEqual(BigInteger.Parse("61290100000000000000000000000"), SunBigUInt10.LeftShiftByBase(612901, 23));
            AssertException<ArgumentOutOfRangeException>(() => SunBigUInt10.LeftShiftByBase(123, -1));

            testHandler.TestGroupEnd();
        }

        public void RunRightShiftByBaseTestCases()
        {
            testHandler.TestGroupStart("Right shift (by base)");

            AssertEqual(123, SunBigUInt10.RightShiftByBase(123, 0));
            AssertEqual(12, SunBigUInt10.RightShiftByBase(123, 1));
            AssertEqual(1, SunBigUInt10.RightShiftByBase(123, 2));
            AssertEqual(0, SunBigUInt10.RightShiftByBase(123, 3));
            AssertEqual(0, SunBigUInt10.RightShiftByBase(123, 43));
            AssertEqual(BigInteger.Parse("612901"),
                        SunBigUInt10.RightShiftByBase(SunBigUInt10.SFromString("61290112345678901234567890123"), 23));

            testHandler.TestGroupEnd();
        }

        public void RunLeftShiftByBinaryTestCases()
        {
            testHandler.TestGroupStart("Left shift (binary)");

            AssertEqual(64, SunBigUInt10.LeftShiftBinary(16, 2));
            AssertEqual(1024, SunBigUInt10.LeftShiftBinary(1, 10));
            AssertEqual(12, SunBigUInt10.LeftShiftBinary(3, 2));
            AssertEqual(3667, SunBigUInt10.LeftShiftBinary(3667, 0));
            AssertEqual(1, SunBigUInt10.LeftShiftBinary(1, 0));
            AssertEqual(0, SunBigUInt10.LeftShiftBinary(0, 34));
            AssertEqual(2097152, SunBigUInt10.LeftShiftBinary(SunBigUInt10.SFromString("1024"), 11));
            AssertEqual(4632491327488, SunBigUInt10.LeftShiftBinary(SunBigUInt10.SFromString("552236"), 23));

            AssertException<ArgumentOutOfRangeException>(() => SunBigUInt10.LeftShiftBinary(123, -1));

            testHandler.TestGroupEnd();
        }

        public void RunRightShiftByBinaryTestCases()
        {
            testHandler.TestGroupStart("Right shift (binary)");
            AssertEqual(16, SunBigUInt10.RightShiftBinary(64, 2));
            AssertEqual(1, SunBigUInt10.RightShiftBinary(1024, 10));
            AssertEqual(3, SunBigUInt10.RightShiftBinary(12, 2));
            AssertEqual(3667, SunBigUInt10.RightShiftBinary(3667, 0));
            AssertEqual(1, SunBigUInt10.RightShiftBinary(1, 0));
            AssertEqual(552236, SunBigUInt10.RightShiftBinary(SunBigUInt10.SFromString("4632491327488"), 23));

            AssertException<ArgumentOutOfRangeException>(() => SunBigUInt10.RightShiftBinary(123, -1));

            testHandler.TestGroupEnd();
        }

        public void RunPowSimpleTestCases()
        {
            testHandler.TestGroupStart("Power (simple)");

            AssertEqual(BigInteger.Pow(BigInteger.Parse("123456789"), 0), SunBigUInt10.PowSimple(SunBigUInt10.SFromString("123456789"), 0));
            AssertEqual(BigInteger.Pow(BigInteger.Parse("123456789"), 1), SunBigUInt10.PowSimple(SunBigUInt10.SFromString("123456789"), 1));
            AssertEqual(BigInteger.Pow(BigInteger.Parse("2"), 5), SunBigUInt10.PowSimple(SunBigUInt10.SFromString("2"), 5));
            AssertEqual(BigInteger.Pow(BigInteger.Parse("123456789"), 5), SunBigUInt10.PowSimple(SunBigUInt10.SFromString("123456789"), 5));
            AssertEqual(BigInteger.Pow(BigInteger.Parse("2"), 10), SunBigUInt10.PowSimple(SunBigUInt10.SFromString("2"), 10));
            AssertEqual(BigInteger.Pow(BigInteger.Parse("2"), 11), SunBigUInt10.PowSimple(SunBigUInt10.SFromString("2"), 11));
            AssertEqual(BigInteger.Pow(BigInteger.Parse("23"), 11), SunBigUInt10.PowSimple(SunBigUInt10.SFromString("23"), 11));
            AssertEqual(BigInteger.Pow(BigInteger.Parse("32327"), 134), SunBigUInt10.PowSimple(SunBigUInt10.SFromString("32327"), 134));

            testHandler.TestGroupEnd();
        }

        private void AssertException<T>(Action a) where T : Exception
        {
            try
            {
                a.Invoke();
                this.testHandler.TestStepFail("No exception occured.");
            }
            catch (T)
            {
                this.testHandler.TestStepSuccess();
            }
            catch (Exception)
            {
                this.testHandler.TestStepFail("Wrong exception occured.");
            }

        }

        private void AssertEqual(BigInteger expectedInt, SunBigUInt10 actualInt)
        {
            string expIntStr = expectedInt.ToString();
            string actIntStr = actualInt.ToString();

            if (expIntStr.Equals(actIntStr))
            {
                this.testHandler.TestStepSuccess();
            }
            else
            {
                this.testHandler.TestStepFail(expectedInt, actualInt);
                NumOfFailedTestcases++;
            }
        }
    }
    }
