﻿using SunBigUInteger.SunBigUInt;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using System.Text;
using System.Threading.Tasks;

namespace SunBigUInteger.SunBigUIntTests
{
    class SunBigUInt256SpecificTests
    {
        private readonly ITestHandler testHandler;
        public int NumOfFailedTestcases { private set; get; }

        public SunBigUInt256SpecificTests(ITestHandler ts)
        {
            this.testHandler = ts;
            NumOfFailedTestcases = 0;
        }

        public void Run()
        {
            RunLeftShiftByBaseTestCases();
            RunRightShiftByBaseTestCases();
            RunLeftShiftByBinaryTestCases();
            RunRightShiftByBinaryTestCases();
        }

        public void RunLeftShiftByBaseTestCases()
        {
            testHandler.TestGroupStart("Left shift (by base)");

            AssertEqual(123, SunBigUInt256.LeftShiftByBase(123, 0));
            AssertEqual(0, SunBigUInt256.LeftShiftByBase(0, 3));
            AssertEqual(31488, SunBigUInt256.LeftShiftByBase(123, 1));
            AssertEqual(BigInteger.Parse("2268949521066274848768"), SunBigUInt256.LeftShiftByBase(123, 8));
            AssertEqual(BigInteger.Parse("15028288791875906354827028020599226383185741227607231316361216"), SunBigUInt256.LeftShiftByBase(612901, 23));
            AssertException<ArgumentOutOfRangeException>(() => SunBigUInt256.LeftShiftByBase(123, -1));

            testHandler.TestGroupEnd();
        }

        public void RunRightShiftByBaseTestCases()
        {
            testHandler.TestGroupStart("Right shift (by base)");

            AssertEqual(123, SunBigUInt256.RightShiftByBase(123, 0));
            AssertEqual(1, SunBigUInt256.RightShiftByBase(257, 1));
            AssertEqual(1, SunBigUInt256.RightShiftByBase(256, 1));
            AssertEqual(0, SunBigUInt256.RightShiftByBase(255, 1));
            AssertEqual(1, SunBigUInt256.RightShiftByBase(65540, 2));
            AssertEqual(0, SunBigUInt256.RightShiftByBase(2589, 3));
            AssertEqual(0, SunBigUInt256.RightShiftByBase(123, 43));
            AssertEqual(BigInteger.Parse("14270216307062399861"),
                        SunBigUInt256.RightShiftByBase(SunBigUInt256.SFromString("61290112345678901234567890123"), 4));

            testHandler.TestGroupEnd();
        }

        public void RunLeftShiftByBinaryTestCases()
        {
            testHandler.TestGroupStart("Left shift (binary)");

            AssertEqual(64, SunBigUInt256.LeftShiftBinary(16, 2));
            AssertEqual(1024, SunBigUInt256.LeftShiftBinary(1, 10));
            AssertEqual(12, SunBigUInt256.LeftShiftBinary(3, 2));
            AssertEqual(3667, SunBigUInt256.LeftShiftBinary(3667, 0));
            AssertEqual(1, SunBigUInt256.LeftShiftBinary(1, 0));
            AssertEqual(0, SunBigUInt256.LeftShiftBinary(0, 34));
            AssertEqual(2097152, SunBigUInt256.LeftShiftBinary(SunBigUInt256.SFromString("1024"), 11));
            AssertEqual(4632491327488, SunBigUInt256.LeftShiftBinary(SunBigUInt256.SFromString("552236"), 23));

            AssertException<ArgumentOutOfRangeException>(() => SunBigUInt256.LeftShiftBinary(123, -1));

            testHandler.TestGroupEnd();
        }

        public void RunRightShiftByBinaryTestCases()
        {
            testHandler.TestGroupStart("Right shift (binary)");
            AssertEqual(16, SunBigUInt256.RightShiftBinary(64, 2));
            AssertEqual(1, SunBigUInt256.RightShiftBinary(1024, 10));
            AssertEqual(3, SunBigUInt256.RightShiftBinary(12, 2));
            AssertEqual(3667, SunBigUInt256.RightShiftBinary(3667, 0));
            AssertEqual(1, SunBigUInt256.RightShiftBinary(1, 0));
            AssertEqual(552236, SunBigUInt256.RightShiftBinary(SunBigUInt256.SFromString("4632491327488"), 23));

            AssertException<ArgumentOutOfRangeException>(() => SunBigUInt256.RightShiftBinary(123, -1));

            testHandler.TestGroupEnd();
        }

        private void AssertException<T>(Action a) where T : Exception
        {
            try
            {
                a.Invoke();
                this.testHandler.TestStepFail("No exception occured.");
            }
            catch (T)
            {
                this.testHandler.TestStepSuccess();
            }
            catch (Exception)
            {
                this.testHandler.TestStepFail("Wrong exception occured.");
            }

        }

        private void AssertEqual(BigInteger expectedInt, SunBigUInt256 actualInt)
        {
            string expIntStr = expectedInt.ToString();
            string actIntStr = actualInt.ToString();

            if (expIntStr.Equals(actIntStr))
            {
                this.testHandler.TestStepSuccess();
            }
            else
            {
                this.testHandler.TestStepFail(expectedInt, actualInt);
                NumOfFailedTestcases++;
            }
        }
    }
}
